﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Collections.ObjectModel;

using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Web.UI;

using Hjg.Pngcs;
using Hjg.Pngcs.Chunks;

using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.Common.Entities;
using Vintagestory.API.Config;
using Vintagestory.API.Datastructures;
using Vintagestory.API.MathTools;
using Vintagestory.Common;

namespace Automap
{
	public class AutomapSystem
	{
		private Thread cartographer_thread;
		private ICoreClientAPI ClientAPI { get; set; }
		private ILogger Logger { get; set; }
		private IChunkRenderer ChunkRenderer { get; set; }

		private const string _mapPath = @"Maps";
		private const string _chunkPath = @"Chunks";
		private const string _domain = @"automap";
		private const string chunkFile_filter = @"*_*.png";
		private static Regex chunkShardRegex = new Regex(@"(?<X>[\d]+)_(?<Z>[\d]+).png", RegexOptions.Singleline);

		private ConcurrentDictionary<Vec2i, uint> columnCounter = new ConcurrentDictionary<Vec2i, uint>(3, 150);
		private ColumnsMetadata chunkTopMetadata;
		private PointsOfInterest POIs = new PointsOfInterest();
		private EntitiesOfInterest EOIs = new EntitiesOfInterest();

		internal Dictionary<int, BlockDesignator> BlockID_Designators { get; private set; }
		internal Dictionary<AssetLocation, EntityDesignator> Entity_Designators { get; private set; }
		internal Dictionary<int, string> RockIdCodes { get; private set; }

		internal RunState CurrentState { get; set; }
		//Run status, Chunks processed, stats, center of map....
		private uint nullChunkCount, updatedChunksTotal;
		private Vec2i startChunkColumn;

		private readonly int chunkSize;
		private string path;
		private IAsset stylesFile;

		public static string AutomapStatusEventKey = @"AutomapStatus";
		public static string AutomapCommandEventKey = @"AutomapCommand";


		public AutomapSystem(ICoreClientAPI clientAPI, ILogger logger)
		{
			this.ClientAPI = clientAPI;
			this.Logger = logger;
			chunkSize = ClientAPI.World.BlockAccessor.ChunkSize;
			ClientAPI.Event.LevelFinalize += EngageAutomap;

			//TODO:Choose which one from GUI 
			this.ChunkRenderer = new StandardRenderer(clientAPI, logger);

			//Listen on bus for commands
			ClientAPI.Event.RegisterEventBusListener(CommandListener, 1.0, AutomapSystem.AutomapCommandEventKey);

		}


		#region Internals
		private void EngageAutomap()
		{
			path = ClientAPI.GetOrCreateDataPath(_mapPath);
			path = ClientAPI.GetOrCreateDataPath(Path.Combine(path, "World_" + ClientAPI.World.Seed));//Add name of World too...'ServerApi.WorldManager.CurrentWorldName'

			stylesFile = ClientAPI.World.AssetManager.Get(new AssetLocation(_domain, "config/automap_format.css"));
			Logger.VerboseDebug("CSS loaded: {0} size: {1}", stylesFile.IsLoaded(), stylesFile.ToText().Length);

			Prefill_POI_Designators();
			startChunkColumn = new Vec2i((ClientAPI.World.Player.Entity.LocalPos.AsBlockPos.X / chunkSize), (ClientAPI.World.Player.Entity.LocalPos.AsBlockPos.Z / chunkSize));
			chunkTopMetadata = new ColumnsMetadata(startChunkColumn);

			Logger.Notification("AUTOMAP Start {0}", startChunkColumn);
			Reload_Metadata();

			ClientAPI.Event.ChunkDirty += ChunkAChanging;

			cartographer_thread = new Thread(Cartographer);
			cartographer_thread.Name = "Cartographer";
			cartographer_thread.Priority = ThreadPriority.Lowest;
			cartographer_thread.IsBackground = true;

			ClientAPI.Event.RegisterGameTickListener(AwakenCartographer, 6000);
		}

		private void ChunkAChanging(Vec3i chunkCoord, IWorldChunk chunk, EnumChunkDirtyReason reason)
		{
			Vec2i topPosition = new Vec2i(chunkCoord.X, chunkCoord.Z);

			columnCounter.AddOrUpdate(topPosition, 1, (key, colAct) => colAct + 1);
		}

		private void AwakenCartographer(float delayed)
		{

			if (CurrentState == RunState.Run && (ClientAPI.IsGamePaused != false || ClientAPI.IsShuttingDown != true))
			{
#if DEBUG
				Logger.VerboseDebug("Cartographer re-trigger from [{0}]", cartographer_thread.ThreadState);
#endif

				if (cartographer_thread.ThreadState.HasFlag(ThreadState.Unstarted))
				{
					cartographer_thread.Start();
				}
				else if (cartographer_thread.ThreadState.HasFlag(ThreadState.WaitSleepJoin))
				{
					//Time to (re)write chunk shards
					cartographer_thread.Interrupt();
				}
				//#if DEBUG
				//ClientAPI.TriggerChatMessage($"Automap {updatedChunksTotal} Updates - MAX (N:{chunkTopMetadata.North_mostChunk},S:{chunkTopMetadata.South_mostChunk},E:{chunkTopMetadata.East_mostChunk}, W:{chunkTopMetadata.West_mostChunk} - TOTAL: {chunkTopMetadata.Count})");
				//#endif
			}
			else if (CurrentState == RunState.Snapshot)
			{
				//TODO: Snapshot generator second thread...
			}

		}


		private void Cartographer()
		{
			wake:
			Logger.VerboseDebug("Cartographer thread awoken");

			try
			{
				uint ejectedItem = 0;
				uint updatedChunks = 0;

				//-- Should dodge enumerator changing underfoot....at a cost.
				if (!columnCounter.IsEmpty)
				{
					var tempSet = columnCounter.ToArray().OrderByDescending(kvp => kvp.Value);
					foreach (var mostActiveCol in tempSet)
					{

						var mapChunk = ClientAPI.World.BlockAccessor.GetMapChunk(mostActiveCol.Key);

						if (mapChunk == null)
						{
							Logger.Warning("SKIP CHUNK: ({0}) - Map Chunk NULL!", mostActiveCol.Key);
							nullChunkCount++;
							columnCounter.TryRemove(mostActiveCol.Key, out ejectedItem);
							continue;
						}

						ColumnMeta chunkMeta = CreateColumnMetadata(mostActiveCol, mapChunk);
						PngWriter pngWriter = SetupPngImage(mostActiveCol.Key, chunkMeta);
						UpdateEntityMetadata();
						ProcessChunkBlocks(mostActiveCol.Key, mapChunk, chunkMeta);

						uint updatedPixels = 0;

						ChunkRenderer.GenerateChunkPngShard(mostActiveCol.Key, mapChunk, chunkMeta, pngWriter, out updatedPixels);

						if (updatedPixels > 0)
						{

#if DEBUG
							Logger.VerboseDebug("Wrote chunk shard: ({0}) - Edits#:{1}, Pixels#:{2}", mostActiveCol.Key, mostActiveCol.Value, updatedPixels);
#endif
							updatedChunks++;
							chunkTopMetadata.Update(chunkMeta);
							columnCounter.TryRemove(mostActiveCol.Key, out ejectedItem);
						}
						else
						{
							columnCounter.TryRemove(mostActiveCol.Key, out ejectedItem);
							Logger.VerboseDebug("Un-painted chunk: ({0}) ", mostActiveCol.Key);
						}

					}
				}

				UpdateStatus(this.updatedChunksTotal, this.nullChunkCount, updatedChunks);

				if (updatedChunks > 0)
				{
					//What about chunk updates themselves; a update bitmap isn't kept...
					updatedChunksTotal += updatedChunks;
					GenerateMapHTML();
					GenerateJSONMetadata();
					updatedChunks = 0;
				}

				//Then sleep until interupted again, and repeat

				Logger.VerboseDebug("Thread '{0}' about to sleep indefinitely.", Thread.CurrentThread.Name);

				Thread.Sleep(Timeout.Infinite);

			}
			catch (ThreadInterruptedException)
			{

				Logger.VerboseDebug("Thread '{0}' interupted [awoken]", Thread.CurrentThread.Name);
				goto wake;

			}
			catch (ThreadAbortException)
			{
				Logger.VerboseDebug("Thread '{0}' aborted.", Thread.CurrentThread.Name);

			}
			finally
			{
				Logger.VerboseDebug("Thread '{0}' executing finally block.", Thread.CurrentThread.Name);
			}
		}

		private void UpdateStatus(uint totalUpdates, uint voidChunks, uint delta)
		{
			StatusData updateData = new StatusData(totalUpdates, voidChunks, delta, RunState.Run);

			this.ClientAPI.Event.PushEvent(AutomapStatusEventKey, updateData);
		}

		private void Prefill_POI_Designators()
		{

			this.BlockID_Designators = new Dictionary<int, BlockDesignator>();
			this.Entity_Designators = new Dictionary<AssetLocation, EntityDesignator>();
			this.RockIdCodes = Helpers.ArbitrarytBlockIdHunter(ClientAPI, new AssetLocation(GlobalConstants.DefaultDomain, "rock"), EnumBlockMaterial.Stone);

			//Add special marker types for BlockID's of "Interest", overwrite colour, and method

			Install_POI_Designators(DefaultDesignators.DefaultBlockDesignators(), DefaultDesignators.DefaultEntityDesignators());
		}

		private void Install_POI_Designators(ICollection<BlockDesignator> blockDesig, List<EntityDesignator> entDesig)
		{
			Logger.VerboseDebug("Connecting {0} standard Block-Designators", blockDesig.Count);
			foreach (var designator in blockDesig)
			{
				var blockIDs = Helpers.ArbitrarytBlockIdHunter(ClientAPI, designator.Pattern, designator.Material);
				if (blockIDs.Count > 0) { Logger.VerboseDebug("Designator {0} has {1} associated blockIDs", designator.ToString(), blockIDs.Count); }
				foreach (var entry in blockIDs)
				{
					BlockID_Designators.Add(entry.Key, designator);
				}
			}
			this.ChunkRenderer.BlockID_Designators = BlockID_Designators;


			Logger.VerboseDebug("Connecting {0} standard Entity-Designators", entDesig.Count);
			foreach (var designator in entDesig)
			{
				//Get Variants first, from EntityTypes...better be populated!
				var matched = ClientAPI.World.EntityTypes.FindAll(entp => entp.Code.BeginsWith(designator.Pattern.Domain, designator.Pattern.Path));

				foreach (var match in matched)
				{
					Logger.VerboseDebug("Linked Entity: {0} Designator: {1}", match.Code, designator);
					this.Entity_Designators.Add(match.Code, designator);
				}



				//EntityProperties props = ClientAPI.World.GetEntityType(designator.Pattern);
			}


		}


		private void GenerateMapHTML()
		{
			string mapFilename = Path.Combine(path, "Automap.html");

			int TopNorth = chunkTopMetadata.North_mostChunk;
			int TopSouth = chunkTopMetadata.South_mostChunk;
			int TopEast = chunkTopMetadata.East_mostChunk;
			int TopWest = chunkTopMetadata.West_mostChunk;

			using (StreamWriter outputText = new StreamWriter(File.Open(mapFilename, FileMode.Create, FileAccess.Write, FileShare.ReadWrite)))
			{
				using (HtmlTextWriter tableWriter = new HtmlTextWriter(outputText))
				{
					tableWriter.BeginRender();
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Html);

					tableWriter.RenderBeginTag(HtmlTextWriterTag.Head);
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Title);
					tableWriter.WriteEncodedText("Generated Automap");
					tableWriter.RenderEndTag();
					//CSS  style  here
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Style);
					tableWriter.Write(stylesFile.ToText());
					tableWriter.RenderEndTag();//</style>

					//## JSON map-state data ######################
					tableWriter.AddAttribute(HtmlTextWriterAttribute.Type, "text/javascript");
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Script);

					tableWriter.Write("var available_images = [");

					foreach (var shard in this.chunkTopMetadata)
					{
						tableWriter.Write("{{X:{0},Y:{1} }}, ", shard.Location.X, shard.Location.Y);
					}

					tableWriter.Write(" ];\n");

					tableWriter.RenderEndTag();

					tableWriter.RenderEndTag();

					tableWriter.RenderBeginTag(HtmlTextWriterTag.Body);
					tableWriter.RenderBeginTag(HtmlTextWriterTag.P);
					tableWriter.WriteEncodedText($"Created {DateTimeOffset.UtcNow.ToString("u")}");
					tableWriter.RenderEndTag();
					tableWriter.RenderBeginTag(HtmlTextWriterTag.P);
					tableWriter.WriteEncodedText($"W:{TopWest}, E: {TopEast}, N:{TopNorth}, S:{TopSouth} ");
					tableWriter.RenderEndTag();
					tableWriter.WriteLine();
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Table);
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Caption);
					tableWriter.WriteEncodedText($"Start: {startChunkColumn}, Seed: {ClientAPI.World.Seed}\n");
					tableWriter.RenderEndTag();

					//################ X-Axis <thead> #######################
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Thead);
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Tr);

					tableWriter.RenderBeginTag(HtmlTextWriterTag.Th);
					tableWriter.Write("N, W");
					tableWriter.RenderEndTag();

					for (int xAxisT = TopWest; xAxisT <= TopEast; xAxisT++)
					{
						tableWriter.RenderBeginTag(HtmlTextWriterTag.Th);
						tableWriter.Write(xAxisT);
						tableWriter.RenderEndTag();
					}

					tableWriter.RenderBeginTag(HtmlTextWriterTag.Th);
					tableWriter.Write("N, E");
					tableWriter.RenderEndTag();

					tableWriter.RenderEndTag();
					tableWriter.RenderEndTag();
					//###### </thead> ################################

					//###### <tbody> - Chunk rows & Y-axis cols
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Tbody);

					//######## <tr> for every vertical row
					for (int yAxis = TopNorth; yAxis <= TopSouth; yAxis++)
					{
						tableWriter.RenderBeginTag(HtmlTextWriterTag.Tr);
						tableWriter.RenderBeginTag(HtmlTextWriterTag.Td);
						tableWriter.Write(yAxis);//legend: Y-axis
						tableWriter.RenderEndTag();

						for (int xAxis = TopWest; xAxis <= TopEast; xAxis++)
						{
							//###### <td>  #### for chunk shard 
							tableWriter.RenderBeginTag(HtmlTextWriterTag.Td);
							var colLoc = new Vec2i(xAxis, yAxis);
							if (chunkTopMetadata.Contains(colLoc))
							{
								ColumnMeta meta = chunkTopMetadata[colLoc];
								//Tooltip first					
								tableWriter.AddAttribute(HtmlTextWriterAttribute.Class, "tooltip");
								tableWriter.RenderBeginTag(HtmlTextWriterTag.Div);

								tableWriter.AddAttribute(HtmlTextWriterAttribute.Src, $"{xAxis}_{yAxis}.png");
								tableWriter.RenderBeginTag(HtmlTextWriterTag.Img);
								tableWriter.RenderEndTag();
								// <span class="tooltiptext">Tooltip text
								tableWriter.AddAttribute(HtmlTextWriterAttribute.Class, "tooltiptext");
								tableWriter.RenderBeginTag(HtmlTextWriterTag.Span);

								StringBuilder tooltipText = new StringBuilder();
								tooltipText.Append($"{meta.Location.PrettyCoords(ClientAPI)} ");
								tooltipText.Append($" Max-Height: {meta.YMax}, Temp: {meta.Temperature.ToString("F1")} ");
								tooltipText.Append($" Rainfall: {meta.Rainfall.ToString("F1")}, ");
								tooltipText.Append($" Shrubs: {meta.ShrubDensity.ToString("F1")}, ");
								tooltipText.Append($" Forest: {meta.ForestDensity.ToString("F1")}, ");
								tooltipText.Append($" Fertility: {meta.Fertility.ToString("F1")}, ");

								if (meta.RockRatio != null)
								{
									foreach (KeyValuePair<int, uint> blockID in meta.RockRatio)
									{
										var block = ClientAPI.World.GetBlock(blockID.Key);
										tooltipText.AppendFormat(" {0} × {1},\t", block.Code.GetName(), meta.RockRatio[blockID.Key]);
									}
								}

								tableWriter.WriteEncodedText(tooltipText.ToString());

								tableWriter.RenderEndTag();//</span>


								tableWriter.RenderEndTag();//</div> --tooltip enclosure
							}
							else
							{
								tableWriter.Write("?");
							}

							tableWriter.RenderEndTag();
						}//############ </td> ###########

						tableWriter.RenderBeginTag(HtmlTextWriterTag.Td);
						tableWriter.Write(yAxis);//legend: Y-axis
						tableWriter.RenderEndTag();

						tableWriter.RenderEndTag();

					}
					tableWriter.RenderEndTag();

					//################ X-Axis <tfoot> #######################
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Tfoot);
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Tr);

					tableWriter.RenderBeginTag(HtmlTextWriterTag.Td);
					tableWriter.Write("S, W");
					tableWriter.RenderEndTag();

					for (int xAxisB = TopWest; xAxisB <= TopEast; xAxisB++)
					{
						tableWriter.RenderBeginTag(HtmlTextWriterTag.Td);
						tableWriter.Write(xAxisB);
						tableWriter.RenderEndTag();
					}

					tableWriter.RenderBeginTag(HtmlTextWriterTag.Td);
					tableWriter.Write("S, E");
					tableWriter.RenderEndTag();

					tableWriter.RenderEndTag();
					tableWriter.RenderEndTag();
					//###### </tfoot> ################################


					tableWriter.RenderEndTag();//</table>

					//############## POI list #####################
					tableWriter.RenderBeginTag(HtmlTextWriterTag.P);
					tableWriter.WriteLine("Points of Interest");
					tableWriter.RenderEndTag();
					tableWriter.RenderBeginTag(HtmlTextWriterTag.Ul);
					foreach (var poi in this.POIs)
					{
						tableWriter.RenderBeginTag(HtmlTextWriterTag.Li);
						tableWriter.WriteEncodedText(poi.Location.PrettyCoords(this.ClientAPI) + "\t");
						tableWriter.WriteEncodedText(poi.Notes + "\t");
						tableWriter.WriteEncodedText(poi.Timestamp.ToString("u"));
						tableWriter.RenderEndTag();
					}

					foreach (var eoi in this.EOIs.PointsList)
					{
						tableWriter.RenderBeginTag(HtmlTextWriterTag.Li);
						tableWriter.WriteEncodedText(eoi.Location.PrettyCoords(this.ClientAPI) + "\t");
						tableWriter.WriteEncodedText(eoi.Notes + "\t");
						tableWriter.WriteEncodedText(eoi.Timestamp.ToString("u"));
						tableWriter.RenderEndTag();
					}

					tableWriter.RenderEndTag();




					tableWriter.RenderEndTag();//### </BODY> ###

					tableWriter.EndRender();
					tableWriter.Flush();
				}
				outputText.Flush();
			}

			Logger.VerboseDebug("Generated HTML map");
		}

		/// <summary>
		/// Generates the JSON Metadata. (in MAP object format )
		/// </summary>
		private void GenerateJSONMetadata()
		{
			string jsonFilename = Path.Combine(path, "Metadata.js");

			StreamWriter jsonWriter = new StreamWriter(jsonFilename, false, Encoding.UTF8);
			using (jsonWriter)
			{
				jsonWriter.WriteLine("var worldSeedNum = {0};", ClientAPI.World.Seed);
				jsonWriter.WriteLine("var genTime = new Date('{0}');", DateTimeOffset.UtcNow.ToString("O"));
				jsonWriter.WriteLine("var startCoords = {{X:{0},Y:{1}}};", startChunkColumn.X, startChunkColumn.Y);
				jsonWriter.WriteLine("var chunkSize = {0};", chunkSize);
				jsonWriter.WriteLine("var northMostChunk ={0};", chunkTopMetadata.North_mostChunk);
				jsonWriter.WriteLine("var southMostChunk ={0};", chunkTopMetadata.South_mostChunk);
				jsonWriter.WriteLine("var eastMostChunk ={0};", chunkTopMetadata.East_mostChunk);
				jsonWriter.WriteLine("var westMostChunk ={0};", chunkTopMetadata.West_mostChunk);
				//MAP object format - [key, value]: key is "x_y"
				jsonWriter.Write("let shardsMetadata = new Map([");
				foreach (var shard in chunkTopMetadata)
				{
					jsonWriter.Write("['{0}_{1}',", shard.Location.X, shard.Location.Y);
					jsonWriter.Write("{");
					jsonWriter.Write("ChunkAge: '{0}',", shard.ChunkAge);//World age - relative? or last edit ??
					jsonWriter.Write("Temperature: {0},", shard.Temperature.ToString("F1"));
					jsonWriter.Write("YMax: {0},", shard.YMax);
					jsonWriter.Write("Fertility: {0},", shard.Fertility.ToString("F1"));
					jsonWriter.Write("ForestDensity: {0},", shard.ForestDensity.ToString("F1"));
					jsonWriter.Write("Rainfall: {0},", shard.Rainfall.ToString("F1"));
					jsonWriter.Write("ShrubDensity: {0},", shard.ShrubDensity.ToString("F1"));
					jsonWriter.Write("AirBlocks: {0},", shard.AirBlocks);
					jsonWriter.Write("NonAirBlocks: {0},", shard.NonAirBlocks);
					//TODO: Heightmap
					//TODO: Rock-ratio
					jsonWriter.Write("}],");
				}
				jsonWriter.Write("]);\n\n");


				jsonWriter.Write("let pointsOfInterest = new Map([");
				foreach (var poi in POIs)
				{
					jsonWriter.Write("['{0}_{1}',", poi.Location.X, poi.Location.Y);
					jsonWriter.Write("{");
					jsonWriter.Write("notes: '{0}',", poi.Notes.Replace("'", " "));
					jsonWriter.Write("timestamp : new Date('{0}'),", poi.Timestamp.ToString("O"));
					jsonWriter.Write("chunkPos:'{0}_{1}',", (poi.Location.X / chunkSize), (poi.Location.Y / chunkSize));
					jsonWriter.Write("}],");
				}

				foreach (var poi in EOIs.PointsList)
				{
					jsonWriter.Write("['{0}_{1}',", poi.Location.X, poi.Location.Y);
					jsonWriter.Write("{");
					jsonWriter.Write("notes: '{0}',", poi.Notes.Replace("'", " "));
					jsonWriter.Write("timestamp : new Date('{0}'),", poi.Timestamp.ToString("O"));
					jsonWriter.Write("chunkPos:'{0}_{1}',", (poi.Location.X / chunkSize), (poi.Location.Y / chunkSize));
					jsonWriter.Write("}],");
				}
				jsonWriter.Write("]);\n\n");

				jsonWriter.Flush();
			}

		}


		private ColumnMeta CreateColumnMetadata(KeyValuePair<Vec2i, uint> mostActiveCol, IMapChunk mapChunk)
		{
			ColumnMeta data = new ColumnMeta(mostActiveCol.Key.Copy(), chunkSize);
			BlockPos equivBP = new BlockPos(mostActiveCol.Key.X * chunkSize,
											mapChunk.YMax,
											mostActiveCol.Key.Y * chunkSize);

			var climate = ClientAPI.World.BlockAccessor.GetClimateAt(equivBP);
			data.UpdateFieldsFrom(climate, mapChunk, TimeSpan.FromHours(ClientAPI.World.Calendar.TotalHours));

			return data;
		}

		/// <summary>
		/// Reload chunk bounds from chunk shards
		/// </summary>
		/// <returns>The metadata.</returns>
		private void Reload_Metadata()
		{
			var worldmapDir = new DirectoryInfo(path);

			if (worldmapDir.Exists)
			{

				var files = worldmapDir.GetFiles(chunkFile_filter);

				if (files.Length > 0)
				{
#if DEBUG
					Logger.VerboseDebug("{0} Existing world chunk shards", files.Length);
#endif



					foreach (var shardFile in files)
					{

						if (shardFile.Length < 512) continue;
						var result = chunkShardRegex.Match(shardFile.Name);
						if (result.Success)
						{
							int X_chunk_pos = int.Parse(result.Groups["X"].Value);
							int Z_chunk_pos = int.Parse(result.Groups["Z"].Value);

							//Parse PNG chunks for METADATA in shard
							using (var fileStream = shardFile.OpenRead())
							{
								//TODO: Add corrupted PNG Exception handing HERE !
								PngReader pngRead = new PngReader(fileStream);
								pngRead.ReadSkippingAllRows();
								pngRead.End();

								PngMetadataChunk metadataFromPng = pngRead.GetChunksList().GetById1(PngMetadataChunk.ID) as PngMetadataChunk;

								chunkTopMetadata.Add(metadataFromPng.ChunkMetadata);
							}

						}
					}

				}
			}
			else
			{
#if DEBUG
				Logger.VerboseDebug("Could not open world map directory");
#endif
			}



		}

		private PngWriter SetupPngImage(Vec2i coord, ColumnMeta metadata)
		{
			ImageInfo imageInf = new ImageInfo(chunkSize, chunkSize, 8, false);

			string filename = $"{coord.X}_{coord.Y}.png";
			filename = Path.Combine(path, filename);

			PngWriter pngWriter = FileHelper.CreatePngWriter(filename, imageInf, true);
			PngMetadata meta = pngWriter.GetMetadata();
			meta.SetTimeNow();
			meta.SetText("Chunk_X", coord.X.ToString("D"));
			meta.SetText("Chunk_Y", coord.Y.ToString("D"));
			//Setup specialized meta-data PNG chunks here...
			PngMetadataChunk pngChunkMeta = new PngMetadataChunk(pngWriter.ImgInfo);
			pngChunkMeta.ChunkMetadata = metadata;
			pngWriter.GetChunksList().Queue(pngChunkMeta);

			return pngWriter;
		}

		/// <summary>
		/// Does the heavy lifting of Scanning columns of chunks - creates Heightmap and Processes POIs, Entities, and stats...
		/// </summary>
		/// <param name="key">Chunk Coordinate</param>
		/// <param name="mapChunk">Map chunk.</param>
		/// <param name="chunkMeta">Chunk metadata</param>
		private void ProcessChunkBlocks(Vec2i key, IMapChunk mapChunk, ColumnMeta chunkMeta)
		{

			int targetChunkY = mapChunk.YMax / chunkSize;//Surface ... 
			for (; targetChunkY > 0; targetChunkY--)
			{
				WorldChunk chunkData = ClientAPI.World.BlockAccessor.GetChunk(key.X, targetChunkY, key.Y) as WorldChunk;

				if (chunkData == null || chunkData.BlockEntities == null)
				{
#if DEBUG
					Logger.VerboseDebug("Chunk null or empty X{0} Y{1} Z{2}", key.X, targetChunkY, key.Y);
#endif
					continue;
				}

				/*************** Chunk Entities Scanning *********************/
				if (chunkData.BlockEntities != null && chunkData.BlockEntities.Length > 0)
				{
#if DEBUG
					Logger.VerboseDebug("Surface@ {0} = BlockEntities: {1}", key, chunkData.BlockEntities.Length);
#endif

					foreach (var blockEnt in chunkData.BlockEntities)
					{

						if (blockEnt != null && blockEnt.Block != null && BlockID_Designators.ContainsKey(blockEnt.Block.BlockId))
						{
							var designator = BlockID_Designators[blockEnt.Block.BlockId];
							designator.SpecialAction(ClientAPI, POIs, blockEnt.Pos.Copy(), blockEnt.Block);
						}
					}

				}
				/********************* Chunk/Column BLOCKs scanning ****************/
				//Heightmap, Stats, block tally
				chunkData.Unpack();

				int X_index, Y_index, Z_index;
				X_index = Y_index = Z_index = 0;

				do
				{
					do
					{
						do
						{
							/* Encode packed indicie
							(y * chunksize + z) * chunksize + x
							*/
							var indicie = Helpers.ChunkBlockIndicie16(X_index, Y_index, Z_index);
							int aBlockId = chunkData.Blocks[indicie];

							if (aBlockId == 0)
							{//Air
								chunkMeta.AirBlocks++;
								continue;
							}

							if (RockIdCodes.ContainsKey(aBlockId))
							{
								if (chunkMeta.RockRatio.ContainsKey(aBlockId)) { chunkMeta.RockRatio[aBlockId]++; } else { chunkMeta.RockRatio.Add(aBlockId, 1); }
							}

							chunkMeta.NonAirBlocks++;

							//Heightmap 
							if (chunkMeta.HeightMap[X_index, Z_index] == 0)
							{ chunkMeta.HeightMap[X_index, Z_index] = (ushort) (Y_index + (targetChunkY * chunkSize)); }

						}
						while (X_index++ < (chunkSize - 1));
						X_index = 0;
					}
					while (Z_index++ < (chunkSize - 1));
					Z_index = 0;
				}
				while (Y_index++ < (chunkSize - 1));

			}
		}

		private void UpdateEntityMetadata()
		{
			Logger.Debug("Presently {0} Entities", ClientAPI.World.LoadedEntities.Count);
			//Mabey scan only for 'new' entities by tracking ID in set?
			foreach (var loadedEntity in ClientAPI.World.LoadedEntities.ToList())
			{

#if DEBUG
				//Logger.VerboseDebug($"ENTITY: ({loadedEntity.Value.Code}) = #{loadedEntity.Value.EntityId} {loadedEntity.Value.State} {loadedEntity.Value.LocalPos}    <<<<<<<<<<<<");
#endif

				var dMatch = Entity_Designators.SingleOrDefault(se => se.Key.Equals(loadedEntity.Value.Code));
				if (dMatch.Value != null)
				{
					dMatch.Value.SpecialAction(ClientAPI, this.EOIs, loadedEntity.Value.LocalPos.AsBlockPos.Copy(), loadedEntity.Value);
				}

			}


		}

		private void AddNote(string notation)
		{
			var playerNodePoi = new PointOfInterest()
			{
				Location = ClientAPI.World.Player.Entity.LocalPos.AsBlockPos.Copy(),
				Notes = notation,
				Timestamp = DateTimeOffset.UtcNow,
			};

			this.POIs.AddReplace(playerNodePoi);
		}



		private void CommandListener(string eventName, ref EnumHandling handling, IAttribute data)
		{
			Logger.VerboseDebug("MsgBus RX: AutomapCommandMsg: {0}", data.ToJsonToken());

			CommandData cmdData = data as CommandData;


			if (CurrentState != RunState.Snapshot)
			{
				switch (cmdData.State)
				{
					case RunState.Run:
						CurrentState = cmdData.State;
						AwakenCartographer(0.0f);
						break;

					case RunState.Stop:
						CurrentState = cmdData.State;
						break;

					case RunState.Snapshot:
						CurrentState = RunState.Stop;
						//Snapshot starts a second thread/process...

						break;

					case RunState.Notation:
						//Add to POI list where player location
						AddNote(cmdData.Notation);
						break;
				}

			}

			if (CurrentState != cmdData.State)
			{
				CurrentState = cmdData.State;
				AwakenCartographer(0.0f);
			}

#if DEBUG
			ClientAPI.TriggerChatMessage($"Automap commanded to: {cmdData.State} ");
#endif

		}


		#endregion



	}

}